	;; shisen 1k, entry for the 2002 minigame competition
	;; by Paolo Ferraris (pieffe8@libero.it) - public domain
	;; works with any ZX Spectrum (16k/48k/128k/etc...) and
	;; most Spectrum emulators - assemblable with TASM


	;; the board is memorized into a table that starts at address
	;; tileaddr. Normally, the address of the tile where the cursor is
	;; on is memorized in HL. L indicates which is the location/space
	;; with the format 0YYYXXXX. L=0 is the top-left corner

dataaddr	.equ	6980h
selected	.equ	0	; selected tile
pairsleft	.equ	1	; pairs left (current)
changed		.equ	2	; 1=board changed, 0=no changes
lastdone	.equ	3	; maximum of pairs left you can go forward
backcolor	.equ	4
gravity		.equ	5	; 255=gravity, 0=no gravity

randsub		.equ	6900h	; address of random subroutine
tileaddr	.equ	6600h	; address of tiles
upptileaddr	.equ	66h
distaddr	.equ	6700h	; address of distances between tiles
uppdistaddr	.equ	67h
prevaddr	.equ	6800h	; address of previous tile in path to tile
deltileaddr	.equ	7000h	; deletion history: type of tile deleted
uppdeltileaddr	.equ	70h
delfirstaddr	.equ	7100h	; deletion history: first tile deleted
uppdelfirstaddr	.equ	71h
delsecondaddr	.equ	7200h	; deletion history: second tile deleted
uppdelsecondaddr .equ	72h
helpaddr	.equ	7300h	; where deletable pairs are memorized
screencolors:	.equ	40	; CYAN paper, BLACK ink
bordercolor:	.equ	5	; CYAN border


	;; the BASIC code
	;; 0 RANDOMIZE: RANDOMIZE RND * USR <start>
	;; 4 INPUT a: RANDOMIZE a: RANDOMIZE USR <newrandnum>: GO TO PI:
	;;   REM <program code>
beginline0:	.ORG	23755
	.byte	0		; line number (higher part)
	.byte	0		; line number (lower part)
	.word	endline0-beginline0-4
	.byte	0F9h		; RANDOMIZE
	.byte	':'		;
	.byte	0F9h		; RANDOMIZE
	.byte	0A5h		; RND
	.byte	'*'
	.byte	0C0h		; USR
	.byte	'0'		; 0
	.byte	0Eh		; number begins
	.byte	0		; 
	.byte	0		; 
	.word	start		; address of start code
	.byte	0
	.byte	13		; ENTER
endline0:
beginline4:
	.byte	0		;
	.byte	4		;
	.word	endline4-beginline4-4
	.byte	238		; INPUT
	.byte	'a'		;
	.byte	':'		;
	.byte	0F9h		; RANDOMIZE
	.byte	'a'		; 
	.byte	':'		;
	.byte	0F9h		; RANDOMIZE
	.byte	0C0h		; USR
	.byte	'0'		; 0
	.byte	0Eh		; number begins
	.byte	0		; 
	.byte	0		; 
	.word	newrandnum	; address of start code
	.byte	0
	.byte	':'
	.byte	0ECh		; GOTO
	.byte	0A7h		; PI
	.byte	':'
	.byte	0EAh		; REM
	;; here the LM code starts
start:
	XOR	A
	LD	(dataaddr+gravity),A	; no gravity, will go to gravity soon
	;; copy the RND math subroutine from ROM to RAM, and append RET.
	LD	HL,025FDh
	LD	DE,randsub
	LD	BC,40
	LDIR
	LD	A,0C9h		; RET
	LD	(DE),A
	;; switches gravity flag
switchgravity:
	LD	HL,dataaddr+gravity
	LD	A,(HL)
	CPL
	LD	(HL),A
restart:
newrandnum:
	LD	IX,dataaddr	; used for most of the bytes
	LD	HL,5C6Ah	; flags2
	SET	3,(HL)		; CAPS LOCK on
	LD	A,screencolors
	LD	(23693),A	; it is easier to POKE than the various calls
	LD	A,bordercolor
	LD	(IX+backcolor),A
	CALL	229Bh		; better than POKE: color change is immediate
	CALL	0D6Bh		; CLS
	LD	BC,(23670)	; seed number
	CALL	2D2Bh		; BC to stack
	CALL	2DE3h		; print that number

	;; random generation of the board
	;; erases the tile board
	LD	HL,tileaddr
loopcleartable:
	LD	(HL),0
	INC	L
	JR	NZ,loopcleartable
	;; fill the board with random tiles (4 per type)
	LD	B,14*6		; number of tiles
loopfillboard:
	;; get a random number from 1 to B inclusive, in C
	LD	A,B
	PUSH	BC
	CALL	02D28h		; A in FP stack
	CALL	randsub		; random number <1 in FP stack
	RST	28h		; do the following operations:
	.byte	4		;   multiplication
	.byte	39		;   INT
	.byte	56		;   end of operations
	CALL	02DA2h		; result from FP stack to A
	POP	BC
	INC	A		; insert
	LD	C,A		; that number in C
	;; now we find the C-th empty space, excluding the first row and
	;; column and the last column
	LD	HL,tileaddr+16	; we start from the second row
loopfindtile:
	INC	L
	LD	A,L
	AND	15		; take only the column number
	DEC	A
	CP	14		; C only if the column is between 1 and 14
	JR	NC,loopfindtile
	LD	A,(HL)
	AND	A
	JR	NZ,loopfindtile	; non empty slot
	DEC	C
	JR	NZ,loopfindtile
	;; we have found that empty space. we put the tile. Each tile type
	;; is a multiple of 8 from 8 to 168 inclusive
	;; and there are four of each type
	LD	A,B		; a unique number from 1 to 84
	ADD	A,3		; a unique number from 4 to 87
	AND	252		; now the number is a multiple of 4
	ADD	A,A
	CALL	inserttile	; we insert (and draw) the tile
	DJNZ	loopfillboard	; repeat until board full
	;; some variable initialization
	LD	HL,2A00h	; number of pairs are 42, selected=0
	LD	(dataaddr+selected),HL
	INC	H		; H=1, board changes
	INC	L		; L=2B, no going forward
	LD	(dataaddr+changed),HL
	LD	HL,tileaddr+11h	; second column, second row
	;; here is the main loop of the program
	;; registers:	HL=address of tile 
mainloop:
	LD	DE,dataaddr+changed
	LD	A,(DE)
	DEC	A		; board changed?
	CALL	Z,findpairs	; if yes, find new pairs and border color
	CALL	showsametile	; show cursor and same tiles
	CALL	getkey		; get a key
	CP	'N'
	JR	Z,restart	; new game
	CP	12
	JP	Z,switchgravity	; new game with gravity switched
	CP	'S'		; new seed?
	RET	Z		; returns to BASIC
	LD	B,A		; save key
	CALL	showsametile	; hide cursor and same tiles
	LD	A,B		; re-get the key
	CALL	handlekeys	; interpret the key
	JR	mainloop	; back to the main loop


	
	;; go back B moves
	;; this procedure is called only when 'b' or 'a' is pressed.
	;; 1 in case of 'b', 42 in case of 'a'
	;; HL=cursor position; HL preserved
undo:
	LD	DE,dataaddr+pairsleft
	LD	A,(DE)
	CP	42		; do not go back if all pairs
	RET	Z		; are on the board
	INC	A
	LD	(DE),A		; increment the number of pairs left
	SET	0,(IX+changed)	; board is changed
	LD	D,uppdelsecondaddr ; (DE) is the board position of the
	LD	E,A		; second element deleted at this step
	CALL	selectcurrres	; unselect the selected tile (if any)
	PUSH	HL		; the cursor position should not change
	LD	A,2
	CALL	changeborder	; red border (cheater!)
	LD	A,(DE)
	LD	C,A		; the second deleted tile
	DEC	D		; (DE)=board position of the first element
	LD	A,(DE)
	LD	L,A		; HL=address of the first deleted tile
	DEC	D		; (DE)=type of deleted tile
	LD	A,(DE)		; the type of tile
	CALL	insertintile	; re-insert the first tile
	LD	A,(DE)		; again, the type of tile
	LD	L,C		; HL=address of the second deleted tile
	CALL	insertintile	; re-insert the second tile
	POP	HL		; restore HL
 	DJNZ	undo		; continue until all back moves are done.
	RET

	;; 'h' has been pressed
	;; HL = cursor position ; HL preserved
askavailablemove:
	LD	BC,helpaddr	; (BC)=first pair of deletable tiles
	LD	D,H
	PUSH	HL		; preserve the cursor position
askavailablemoveloop:
	LD	A,(BC)		; first element
	AND	A
	JR	Z,askavailablemoveexit ; if 0, no other pairs to be shown
	LD	L,A		; that element in E
	INC	BC
	LD	A,(BC)
	LD	E,A		; second element
	INC	BC
	LD	A,(IX+selected)	; selected tile
	AND	A		; if no tile is selected, any pair is
	JR	Z,pairvalid	; good
	CP	L		; we show the pair only if one ot the 
	JR	Z,pairvalid	; two tiles is the one selected
	CP	E		; 
	JR	NZ,askavailablemoveloop	; if not, take next pair
pairvalid:
	CALL	flashpair	; show the pair
	CALL	getkey		; get a key
	PUSH	AF
	CALL	flashpair	; un-show the pair
	POP	AF
	CP	13		; ENTER pressed?
	JR	Z,removepairlong; remove the pair
	CP	'H'		; 'h' pressed again?
	JR	Z,askavailablemoveloop ; show another pair
askavailablemoveexit:
	POP	HL
	RET

	;; handles the key pressed
	;; A=key pressed, HL=cursor position; HL preserved
handlekeys:
	LD	B,1		; one step
	CP	'B'
	JR	Z,undo		; back
	CP	'F'
	JR	Z,redo		; forward
	DEC	B		; B=0 -> at most 256 times - enough
	CP	'L'
	JR	Z,redo		; last step
	CP	'R'
	JR	Z,undo		; initial board
#ifdef	DEBUG
	CP	'D'
	JP	Z,showproxmatr
#endif
	CP	'H'
	JR	Z,askavailablemove
	SUB	7
	LD	B,A		; B=keypressed-7
	LD	A,L		; curpos in A
	DJNZ	contwithright	; not LEFT
	DEC	A		; move left
contwithright:
	DJNZ	testhorizontal	; not RIGHT
	INC	A		; move right
testhorizontal:
	XOR	L
	LD	C,16		; we use that variable several times
	CP	C		; has the Y value changed?
	RET	NC		; if yes, invalid key. and return
	XOR	L		; new curpos again in A
	DJNZ	contwithup	; not DOWN
	ADD	A,C		; go down (add 16)
contwithup:
	DJNZ	testvertical	; not UP
	SUB	C		; go up (subtract 16)
testvertical:
	CP	128		; outside board?
	RET	NC		; if yes, wrong key pressed, return
	LD	L,A		; if no, the new curpos is the final one
	DEC	B		; BACKSPACE pressed?
	DEC	B		; ENTER pressed?
	RET	NZ		; if no, we are done
	;; if there is no tile under the cursor, return
	LD	A,(HL)
	AND	A
	RET	Z
	;; has the tile the same position of the selected one?
	;; if yes, deselect the tile and return.
	;; (note L cannot be 0, as there is no tile at that position)
	LD	A,(IX+selected)
	CP	L
	JR	Z,selectcurrres
	;; is a tile selected? If not, select this one and return
	AND	A
	JR	Z,selectcurrset
	;; is there a path between the two tiles? If not, return
	LD	E,A
	CALL	genproxmatrices
	CALL	checkpath
	RET	NZ
	PUSH	HL		; save the current position
	;; the value in L must be lower than in E. exchange if necessary
	LD	A,E
	CP	L
	JR	C,removepairlong
	EX	DE,HL
	;; this procedure is used if we removed a pair if manually
	;; or with the help (not forward command)
	;; the stack contains the cursor position (full address)
	;; H=upptileaddr, L,E=tiles to be removed (with E>L)
	;; everything is destroyed, except that a POP HL is done at the
	;; end.
removepairlong:
	CALL	drawsegments	; draw the segments
	LD	A,(IX+backcolor)
	LD	B,254
beeploop2:
	PUSH	BC
	OUT	(254),A
	XOR	16
beeploop1:
	NOP
	DJNZ	beeploop1
	POP	BC
	DJNZ	beeploop2
; 	LD	B,10
; waitloop:
; 	HALT
; 	DJNZ	waitloop	; wait a 1/10th of second
	CALL	drawsegments	; delete the segments
	;; updates some data: the pair is putted in the history
	;; we update lastdone. Here everything is destroyed, but
	;; POP HL takes the curpos back.
	LD	C,(HL)
	LD	B,L
	LD	H,uppdeltileaddr
	LD	L,(IX+pairsleft)
	LD	(IX+lastdone),L
	LD	(HL),C		; the tile type
	INC	H
	LD	(HL),B		; the lowest tile on the screen
	INC	H
	LD	(HL),E
	POP	HL		; note:	here B is greater than 0
	;; used when a pair of tiles has to be physically deleted from the
	;; board, at the first deletion or with forward ('f', 'l' keys)
	;; B=number of steps to go on, HL=current position, HL preserved
redo:
	LD	A,(IX+pairsleft)
	CP	(IX+lastdone)	; tiles to be deleted?
	RET	C		; if not, return
	DEC	(IX+pairsleft)	; decrease the number of pairs of tiles
	SET	0,(IX+changed)	; board changes
	PUSH	HL		; preserve HL
	LD	D,uppdelsecondaddr
	LD	E,A		; (DE)=second tile to be deleted
	CALL	selectcurrres	; deselect the selected tile (if any)
	LD	A,(DE)
	LD	L,A		; L=second tile to be deleted
	CALL	removeintile	; remove that tile
	DEC	D
	LD	A,(DE)
	LD	L,A		; L=first tile to be deleted
	CALL	removeintile	; remove that tile
	POP	HL
	DJNZ	redo		; repeat B times
	RET

	;; deselect a selected tile
	;; preserves HL,DE,BC
selectcurrres:
	PUSH	HL
	XOR	A
	LD	L,(IX+selected)	; HL address of tile to be deleted
	CP	L		; is a block selected?
	CALL	NZ,selectandcolor ; if yes, deselect it (A=0)
	POP	HL
	RET

	;; select a tile
	;; HL=address of block to be selected
	;; preserves HL,DE,BC
selectcurrset:
	LD	A,L
selectandcolor:
	LD	(IX+selected),A	; set the variable
	LD	A,32
	JR	xorattrtile	; show the selection on the screen

	;; move a sequence of tiles with the following criteria:
	;; H must contain upptileaddr, the block L+B is moved in L,
	;; L+B+B in L+B, etc... The last moved block is C.
	;; at the end, L=C. Preserves DE,BC,H. If no gravity,
	;; it does nothing.
movetiles:
	LD	A,L
	SUB	C		; if L=C, return
	AND	(IX+gravity)	; if no gravity, return too!
	RET	Z
	LD	A,L
	ADD	A,B
	PUSH	DE
	LD	D,H		; DE=HL+B (note, B has to be considered
	LD	E,A		; between -128,+127)
	LD	A,(DE)		; the type of block to be moved
	CALL	inserttile	; we create a block of that type
	LD	L,E		; HL=HL+B
	CALL	removetile	; we remove this block
	POP	DE
	JR	movetiles

	;; we insert a tile, considering gravity
	;; HL=address of tile to be inserted, A=type of tile to
	;; be inserted. Preserves BC,HL,DE.
insertintile:
	PUSH	BC
	PUSH	AF
	LD	C,L		; we move L and all tiles above it up
	LD	A,L		; C=L is the space that will remain empty
	AND	15		; 
	LD	L,A		; space of the top row
	LD	B,16		; X+B=tile below X
	CALL	movetiles	; move the sequence, (at the end, L has the
				; initial value)
	LD	L,C
	POP	AF		; the type of tile
	POP	BC
	;; used when we create a new tile, when we move blocks, or we
	;; put deleted tile back
	;; A=value of the tile, HL=address of tile to be putted back
	;; preserves A,BC,DE,HL
inserttile:
	LD	(HL),A
	;; used when we put or delete a tile from the screen
	;; HL=address of tile. Tile must exist, otherwise it does
	;; nothing. Preserves BC,DE,HL.
drawandcolortile:
	LD	A,(HL)		; A=type of tile
	AND	A
	RET	Z		; return if no tile
	PUSH	BC
	PUSH	DE
 	PUSH	HL
	PUSH	AF
	CALL	gettilecoords	; DE=top-left coordinate of the tile
	;; draw the frame
	INC	D		; one pixel down
 	LD	HL,tileframe	; address of the tile frame
	CALL	printcolumn23	; draw the rightmost part of the tile frame
	SET	3,E		; adds 8
 	CALL	printcolumn23	; the rightmost part
	;; increase D by 6, and decrease L by 5
 	LD	BC,05FBh	; 
 	EX	DE,HL
 	ADD	HL,BC
 	EX	DE,HL
	;; write the letter in the center of tile
	POP	AF		; the value of the tile
	LD	L,A
	LD	H,3Eh		; HL=first byte of the letter in ROM
	LD	B,8		; number of lines
	CALL	printcolumn	; write the letter
	LD	A,16		; value to be xor-ed to the attributes
 	POP	HL
	JR	xorattrtileafterpush ; call to change the attributes

	;; make two tiles (whose addresses are in DE and HL) flash
	;; preserves DE,HL,BC
flashpair:
	EX	DE,HL
	CALL	flashtile
	EX	DE,HL
flashtile:
	LD	A,128		; flash attribute
	;; xor the value of the attributes on screen of the tile
	;; whose value is in HL with the value in A.
	;; Preserves BC,DE,HL
xorattrtile:
	PUSH	BC
	PUSH	DE
	;; catt to xorattrtile, but after already having DE and BC in stack
xorattrtileafterpush:
	PUSH	HL
	LD	C,A		; C=xor value
	;; get the memory address of left-most and up-most attribute
	LD	H,44		; HL=00101000-0YYYXXXX
 	LD	A,L		; 0YYYXXXX
 	AND	112		; 0YYY0000
 	ADD	A,A		; YYY00000
 	LD	E,A
 	LD	D,0		; DE=00000000-YYY00000
	ADD	HL,DE		; HL=0010100Z-ZZZZXXXX (ZZZZZ=YYY*3)
	ADD	HL,HL		; HL=010100ZZ-ZZZXXXX0
	 			; HL=attribute address+X*2+Y*96
	LD	B,3		; three lines
loopattrtile:
	;; xor the attributes in current line
	LD	A,C
	XOR	(HL)
	LD	(HL),A
	INC	HL
	LD	A,C
	XOR	(HL)
	LD	(HL),A
	LD	DE,31
	ADD	HL,DE		; first attribute in next line
	DJNZ	loopattrtile	; repeat for all lines
	POP	HL
	POP	DE
	POP	BC
	RET


	;; get the coordinates of 3 pixel left of the top-left pixel of
	;; the dot in the center of a tile whose address is in HL. The
	;; coordinates go in DE. Preserves HL.
getdotcoordinates:
	CALL	gettilecoords	; the leftmost-upmost coordinates of tile
	LD	BC,0B04h
	EX	DE,HL
	ADD	HL,BC		; X=X+4, Y=Y+10
	EX	DE,HL
	RET

	;; draw the segments that goes from D to H
	;; H=uppaddrtile. Preserves DE and HL
drawsegments:
	CALL	genproxmatrices	; generate the matrices for H
	PUSH	HL
	PUSH	DE
	LD	L,E		; we start from the last tile
	CALL	getdotcoordinates ; calculate the center (in DE) and
	CALL	drawbigdot	; draw the point
loopdrawsegment1:
	INC	H		; (HL)=distance, in segments, from the
	LD	A,(HL)		; other end
	AND	A		; if it is 0, we are arrived to that end.
	JR	Z,exitdrawsegments
	INC	H		; (HL)=ending point of next segment
	LD	L,(HL)
	DEC	H
	DEC	H		; HL=address of next tile (space if absent)
	PUSH	DE		; current point in stack
	CALL	getdotcoordinates ; get coordinates of new point (in DE)
	PUSH	DE		; new point in stack
	POP	BC		; BC=destination point
	POP	DE		; DE=current point
	;; now we will move point DE toward BC, drawing dots in the path
loopdrawsegment2:
	;; the following lines are like:	CP DE,BC
	EX	DE,HL
	AND	A		; reset C flag
	PUSH	HL
	SBC	HL,BC
	POP	HL
	EX	DE,HL
	JR	Z,loopdrawsegment1 ; we are at the end point of the segment
	PUSH	HL		; preserves the address of the next tile
	;; if DE>BC, we mean that we will have to subtract 2 from
	;; D or E, add 2 otherwise. 2 because it's the size of a dot,
	;; so that a linked sequence of dots are a segment.
	LD	H,2
	JR	C,todecrease	; is BC greater than DE? Then H=2
	LD	H,254		; H=-2, otherwise
todecrease:
	LD	A,D
	CP	B		; is the y coordinate the same?
	JR	Z,othercoordinate
	ADD	A,H		; if no, D is the variable to change
	LD	D,A
	JR	gotoprintdot
othercoordinate:
	LD	A,E		; if yes, the variable to change is E
	ADD	A,H
	LD	E,A
gotoprintdot:
	POP	HL		; re-get the address of the end of the segment
	CALL	drawbigdot	; draw the dot
	JR	loopdrawsegment2; go on (until all segments are done)
exitdrawsegments
	POP	DE
	POP	HL
	RET

	;; draw a 4-pixel dot 3 pixel right of DE
	;; preserves HL,BC,DE
drawbigdot:
	PUSH	HL
	PUSH	BC
	LD	HL,3D75h	; (HL)=(HL+1)=00011000b
	LD	B,2		; exactly two lines to draw
	CALL	printcolumn	; draw these two lines
	POP	BC
	POP	HL
	RET

	;; remove a tile whose address is in HL, and, if gravity, the
	;; above tiles fall down. Preserves HL,BC,DE.
removeintile:
	CALL	removetile	; remove that tile
 	PUSH	HL
	PUSH	BC
	LD	B,0F0h		; -16: L-16=tile just above L
	LD	A,L
	AND	15
	LD	C,A		; C=last block to be moved down. Top row.
	CALL	movetiles	; shifts the tiles down from C to L.
	POP	BC
	POP	HL
	RET

	;; removes a tile whose address is in HL, without considering gravity.
	;; preserves HL,BC,DE
removetile:
	CALL	drawandcolortile; remove it from screeen
	LD	(HL),0		; remove it from memory
	RET

	;; gets the coordinates of the tile L. The coordinates go to DE
	;; preserves HL,BC,AF
gettilecoords:
	PUSH	AF
	LD	A,L
	AND	240		; the y coordinate times 16
	LD	D,A		; in D
	RRCA			; the y coordinate times 8
	ADD	A,D		; the y coordinate times 24
	LD	D,A
	LD	A,L
	AND	15		; the X coordinate
	ADD	A,A
	ADD	A,A
	ADD	A,A
	ADD	A,A		; X times 16
	LD	E,A
	POP	AF
	RET

	;; generates the proximity matrices from tile whose address is in HL
	;; preserves all variables
	;; there are two matrices. The distance one, indicates how many
	;; segments are needed to go to L (0 in L, 255 if >3)
	;; The previous one indicates which is the next tile in the path
	;; to go to L (only if distance is between 1 and 3, inclusive).
	;;
	;; The generation consists in an initialization (the distance all
	;; to 255, except for L), and in three scans of all tiles, each
	;; associated with a value of B that goes from 0 to 2. For each
	;;  tile that has distance B from L, we look at all spaces in the
	;; same row and column that is not hidden by a tile, and these
	;; spaces will have distance B+1, if no better distance has been
	;; found.
genproxmatrices:
	;; initialization
	PUSH	HL
	PUSH	DE
	PUSH	BC
	LD	B,255
	INC	H
	;; all spaces have distance 255
	LD	A,L
	LD	L,B
loopdist:
	LD	(HL),B
	DEC	L
	JR	NZ,loopdist
	LD	L,A
	INC	(HL)		; only (HL) has distance 0
	LD	L,B
nextlevel:
	INC	B		; B now has value 1 or 2
nexttile:
	;; check that the space is empty and distance B
	;; normally (HL)-B is 0 or has bits 2.3.4.5.6.7 high
	;; if we XOR (HL) with ((H-1)H), and ((H-1)H)!=0, we change
	;; one of these bits, so that if ((H-1)H)!=0, the result
	;; of subtraction is !=0 for sure.
	XOR	A
	SUB	B
	DEC	H
	AND	(HL)		; tile type
	INC	H
	XOR	(HL)		; distance
	CP	B
	CALL	Z,goodtile	; if zero, consider that space/tile
	DEC	L
	JR	NZ,nexttile
	BIT	1,B		; is B=2?
	JR	Z,nextlevel	; if no, we can continue
	POP	BC
	POP	DE
	POP	HL
	RET

	;; debug
#ifdef	DEBUG
showproxmatr:
	PUSH	HL
 	PUSH	DE
	PUSH	BC
	PUSH	AF
	CALL	genproxmatrices
	LD	E,127
loopdebug:
	LD	D,upptileaddr
	LD	H,D
	LD	A,(DE)
	LD	B,A
	INC	D
	LD	A,(DE)
	PUSH	HL
	PUSH	DE
	PUSH	BC
	PUSH	AF
	LD	L,E
	LD	A,L
	AND	240
	LD	D,A
	RRCA
	ADD	A,D
	LD	D,A
	LD	A,L
	AND	15
	ADD	A,A		; A=A*16+4
	SCF
	ADC	A,A
	ADD	A,A
	ADD	A,A
	LD	E,A
	POP	AF
	BIT	7,A
	JR	NZ,skipwrite1
	ADD	A,'0'
	CALL	printchar
skipwrite1:
	POP	BC
	LD	A,D
	ADD	A,16
	LD	D,A
	LD	A,B
	AND	A
	JR	Z,skipwrite2
	RRCA
	RRCA
	RRCA
	ADD	A,64
	CALL	printchar
skipwrite2:
	POP	DE
	POP	HL
	DEC	E
	JR	NZ,loopdebug
	POP	AF
	POP	BC
	POP	DE
	POP	HL
	RET
#endif

	;; using the matrices generated for the tile whose address is
	;; in HL, check if L and E can be removed. Z if yes, NZ otherwise.
	;; We suppose that there is a tile in L, and L <> E
	;; 
	;; preserves HL, E,BC. D gets the value of L
checkpath:
	LD	D,H
	LD	A,(DE)
	CP	(HL)		; is there the same type of tyle
	RET	NZ		; if not, exit
	INC	D		; H=uppdistaddr
	LD	A,(DE)		; distance of E
	DEC	D
	BIT	7,A		; is it infinite?
	RET

	;; called by the main loop when the board changes, it generates
	;; the pairs of tiles that can be removed. Changes the border color.
	;; DE=dataaddr+changed, A=0
	;; preserves HL
	;; these pairs of tiles are memorized starting from helpaddr and
	;; are terminated by a 0.
findpairs:
	LD	(DE),A		; write that board not changed
	;; here is where the pairs are found
	PUSH	HL
	LD	BC,helpaddr	; beginning of list
	LD	L,110		; last piece
loopaskav
	LD	A,(HL)
	AND	A		; is that space empty?
	JR	Z,endloopaskav	; if yes, go to next tile
	CALL	genproxmatrices	; generate matrix for that tile
	LD	E,L		; the second tile will have a lower value
loopaskav2:
	DEC	E
	JR	Z,endloopaskav	; if E=0, we have considered all second pairs
	CALL	checkpath	; check a path from L to E
	JR	NZ,loopaskav2	; if no path, check other pairs
	;; found pair!
	LD	A,L
	LD	(BC),A		; first element memorized
	INC	BC
	LD	A,E
	LD	(BC),A		; second element memorized
	INC	BC
	JR	loopaskav2	; check other pairs
endloopaskav:
	DEC	L
	JR	NZ,loopaskav	; have we considered all initial tiles?
	XOR	A
	LD	(BC),A		; put the terminating 0.
	POP	HL
	;; check if there are elements in the list
	OR	C		; A=0, then check if C=0
	RET	NZ		; at least one pair deletable
	CP	(IX+pairsleft)	; A is 0, then we see if (IX+pairsleft)=0
	JR	NZ,changeborder	; if pairs left, the border color is black
	LD	A,7		; white color
changeborder:
	LD	(IX+backcolor),A; change the border color
	OUT	(254),A
	RET

	;; highlight the cursor (in HL) and show tiles of the same type 
	;; preserves HL,DE,B
showsametile:
	LD	A,64		; bright color
	CALL	xorattrtile
	LD	A,(HL)
	AND	A
	RET	Z
	LD	C,(HL)
	PUSH	HL
	LD	L,128
showtileloop:
	LD	A,(HL)
	CP	C
	LD	A,1		; blue ink
	CALL	Z,xorattrtile
	DEC	L
	JR	NZ,showtileloop
	POP	HL
	RET

	;; now we go in the four directions. We will find next space
	;; just adding D to the current one. A will be last space to
	;; consider (if no obstacles before)
goodtile:
	;; UP
	LD	D,0F0h		; -16
	LD	A,L
	AND	15		; top row
	CALL	doall2
	;; DOWN
	LD	D,010h		; +16
	LD	A,L
	OR	112		; bottom row
	CALL	doall2
	;; LEFT
	LD	D,0FFh		; -1
	LD	A,L
	AND	112		; left-most column
	CALL	doall2
	;; RIGHT
	LD	D,001h		; +1
	LD	A,L		; right-most column
	OR	15
	;; this function is used when calculating the matrices
	;; given: H=uppdistaddr, L=starting space
	;; A=ending space, D=such that L+D is next tile
	;; it writes, for each space with infinite distance on that
	;; direction not hidden by a tile, the distance value B+1,
	;; and the preceding space value L.
	;; preserves HL,B
doall2:
	LD	E,A		; memorize the ending space
	LD	C,L		; preserve the starting space
loopdoall2:
	LD	A,L
	CP	E		; ending space?
	JR	Z,exitdoall2	; if yes, go to exit
	ADD	A,D
	LD	L,A		; new space
	LD	A,(HL)
	INC	A		; is distance=255
	JR	NZ,doallskip2	; if not, value already inserted, skip it
	LD	(HL),B
	INC	(HL)		; new distance = B+1
	INC	H
	LD	(HL),C		; preceding tile is memorized
	DEC	H
doallskip2:
	DEC	H
	LD	A,(HL)		; is that space occupied by a tile?
	INC	H
	AND	A
	JR	Z,loopdoall2	; if not, we continue (with next space)
exitdoall2:
	LD	L,C		; reget the original value of L
	RET

#ifdef	DEBUG
printchar:
	PUSH	BC
	PUSH	HL
	PUSH	AF
	;; taken from rom
	LD	BC,3C00h	; 256 less than the character set address
	LD	H,0
	LD	L,A		; the character
	ADD	HL,HL
        ADD	HL,HL
        ADD	HL,HL
        ADD	HL,BC		; the first byte of character
	LD	B,8		; number of lines
	CALL	printcolumn
	POP	AF
	POP	HL
	POP	BC
	RET
#endif

printcolumn23:
	LD	B,23
	;; xor a 8xB matrix on the screen, starting from pixel DE
	;; line Y is determined by byte (HL+Y)
	;; Preserves DE. At the end B=0, HL<-HL+B
printcolumn:
	PUSH	DE		; preserving DE for the end
loopprintcolumn:
	PUSH	HL
	PUSH	DE
	PUSH	BC
	LD	L,(HL)		; we place the byte in L, 0 in H,
	LD	H,0		;   and we will shift it left many times
	PUSH	HL
	LD	C,E
	LD	A,D
	CALL	22B0h		; in ROM. Input: C=xcord, A=ycoord
				; Output: HL screen address, A=byte position
				; destroys: A,BC,HL
	EX	DE,HL		; DE <- address of two bytes in screen
	POP	HL		; HL <- the byte to write
printbyteloop:
	ADD	HL,HL		; left shifting 8-A times
	INC	A
	CP	8
	JR	NZ,printbyteloop
	;; now we can exor H and L in (DE) and (DE+1), respectively
	LD	A,(DE)
	XOR	H
	LD	(DE),A
	INC	DE
	LD	A,(DE)
	XOR	L
	LD	(DE),A
	POP	BC
	POP	DE
	POP	HL
	INC	D		; increases the row
	INC	HL		; increases the byte
	DJNZ	loopprintcolumn	; repeat until all bytes are done
	POP	DE
	RET

	;; wait until a key is pressed
	;; the key pressed goes to A, preserves BC,DE,HL
getkey:
	PUSH	HL
	LD	HL,05C3Bh	; FLAGS
loopgetkey:
	BIT	5,(HL)
	JR	Z,loopgetkey
	LD	A,(05C08h)	; get the char
	RES	5,(HL)
	POP	HL
	RET

tileframe:
	;; tile frame, 16x23
	;; left column
	.byte	01111111b
	.byte	01000000b
	.byte	01000000b
	.byte	01000000b
	.byte	01000000b
	.byte	01000000b
	.byte	01000000b
	.byte	01000000b
	.byte	01000000b
	.byte	01000000b
	.byte	01000000b
	.byte	01000000b
	.byte	01000000b
	.byte	01000000b
	.byte	01000000b
	.byte	01000000b
	.byte	01000000b
	.byte	01000000b
	.byte	01000000b
	.byte	01000000b
	.byte	01111111b
	.byte	00101010b
	.byte	00011111b
	;; right column
	.byte	11111100b
	.byte	00000110b
	.byte	00000101b
	.byte	00000111b
	.byte	00000101b
	.byte	00000111b
	.byte	00000101b
	.byte	00000111b
	.byte	00000101b
	.byte	00000111b
	.byte	00000101b
	.byte	00000111b
	.byte	00000101b
	.byte	00000111b
	.byte	00000101b
	.byte	00000111b
	.byte	00000101b
	.byte	00000111b
	.byte	00000101b
	.byte	00000111b
	.byte	11111101b
	.byte	10101011b
	.byte	11111111b

	.byte	13		; end BASIC line
endline4:
.end
